__author__  = "Marco Musy"
__license__ = "MIT" 
__maintainer__ = "M. Musy"
__email__   = "marco.musy@gmail.com"
__status__  = "dev"
__website__ = "https://github.com/marcomusy/pianoplayer"


from pianoplayer.hand import Hand
from pianoplayer.scorereader import  reader
from pianoplayer.vtk_keyboard import VirtualKeyboard


import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#------------------------------------------------------------------------------- 
# OBSOLETE
#-------------------------------------------------------------------------------

from music21.midi.realtime import StreamPlayer
from music21.stream import Stream
try:
    import visual as vp
    from visual.text import text
except:
    # import vpython as vp
    # from vpython.text import text
    print "VirtualKeyboard: cannot find some package."
    print 'python visual not installed?'
    quit()


version = 'PIANOFING V4.2'


###########################################################
class VirtualKeyboard:

    def __init__(self):
                  
        self.rate = 100
        
        self.KB = dict()
        self.scene = None
        
        self.rightHand = None
        self.leftHand  = None
                
        self.playsounds = False

        self.build_keyboard()
        self.vpRH=None
        self.vpLH=None

    
    #######################################################
    def build_RH(self, hand):
    
        print 'Building Right Hand..'
        self.rightHand = hand
        sz = (0.8, 1, 5)
        f1 = vp.ellipsoid(pos=(-2.,1.5,6.2),size=sz, opacity=0.9, color=(.7,0.3,0.3))
        f2 = vp.ellipsoid(pos=(-1.,1.5,4.2),size=sz, opacity=0.9, color=(.7,0.3,0.3))
        f3 = vp.ellipsoid(pos=( 0.,1.5,4),  size=sz, opacity=0.9, color=(.7,0.3,0.3))
        f4 = vp.ellipsoid(pos=( 1.,1.5,4.1),size=sz, opacity=0.9, color=(.7,0.3,0.3))
        f5 = vp.ellipsoid(pos=( 2.,1.5,4.5),size=sz, opacity=0.9, color=(.7,0.3,0.3))

        palm = vp.ellipsoid(pos=(0,1.6,10), size=(9.6,2.,6),  opacity=0.3, color=(.5,0,0))
        palm.rotate(angle=0.1)
        palm.rotate(angle=0.1, axis=(0,1,0))

        self.vpRH = [palm, f1,f2,f3,f4,f5]
        for limb in self.vpRH: 
            limb.x   *= 2.5
            limb.pos += (16.5*5.+1, 0, 0)
    
    #######################################################
    def build_LH(self, hand):
    
        print 'Building Left Hand..'
        self.leftHand = hand
        sz = (0.8, 1, 5)
        f1 = vp.ellipsoid(pos=(-2.,1.5,6.2),size=sz, opacity=0.9, color=(.7,0.3,0.3))
        f2 = vp.ellipsoid(pos=(-1.,1.5,4.2),size=sz, opacity=0.9, color=(.7,0.3,0.3))
        f3 = vp.ellipsoid(pos=( 0.,1.5,4),  size=sz, opacity=0.9, color=(.7,0.3,0.3))
        f4 = vp.ellipsoid(pos=( 1.,1.5,4.1),size=sz, opacity=0.9, color=(.7,0.3,0.3))
        f5 = vp.ellipsoid(pos=( 2.,1.5,4.5),size=sz, opacity=0.9, color=(.7,0.3,0.3))

        palm = vp.ellipsoid(pos=(0,1.6,10), size=(9.6,2.,6),  opacity=0.3, color=(.5,0,0))
        palm.rotate(angle=-0.1)
        palm.rotate(angle=-0.1, axis=(0,1,0))

        self.vpLH = [palm, f1,f2,f3,f4,f5]
        for limb in self.vpLH: 
            limb.x   *= -2.5
            limb.pos += (16.5*3.+1, 0, 0)
            

    #######################################################
    def build_keyboard(self):
        
        print 'Building Keyboard..'
        nts = ("C","D","E","F","G","A","B")
        tol = 0.12
        keybsize = 16.5 #cm, span of one octave
        wb = keybsize/7.
        nr_octaves = 7
        span = nr_octaves*wb*7.
    
        self.scene = vp.display(title='Piano Keyboard', x=0, y=0, width=1400./1., height=600./1.,  
                                center=(75,0,0), forward=(0.,-2,-1.), background=(0., 0.25, 0.0))
        #wooden top and base
        vp.box(pos=(span/2+keybsize,-1.,-3),length=span+1, height=1, width=17, material=vp.materials.wood)
        vp.box(pos=(span/2+keybsize, 1, -8),length=span+1, height=3, width=7,  material=vp.materials.wood)
        text(pos=(28,2.2,-8), string=version, width=2, height=2, up=(0,0,-1),
             color=vp.color.orange, depth=0.3, justify='center')
        #leggio
        leggio = vp.box(pos=(75, 8., -12.), length=span/2, height=span/8, width=0.08, color=(1,1,0.9))
        leggio.rotate(angle=-0.4)

        for ioct in range(nr_octaves):
            for ik in range(7):#white keys
                x  = ik * wb + (ioct+1.)*keybsize +wb/2.
                tb = vp.box(pos=(x,0.,0), length=wb-tol, height=1, width=10, up=(0,1,0), color=(1,1,1))
                self.KB.update({nts[ik]+str(ioct+1) : tb})
                if not nts[ik] in ("E","B"): #black keys
                    tn=vp.box(pos=(x+wb/2,wb/2,-2), length=wb*.6, height=1, width=6, up=(0,1,0), color=(0,0,0))
                    self.KB.update({nts[ik]+"#"+str(ioct+1) : tn})

        self.scene.lights = []
        vp.local_light(pos=(  0, 100,  0), color=vp.color.white) #source1
        vp.local_light(pos=(-10, -40, 20), color=vp.color.white) #source2


    ####################################################### 
    def playsound(self, n):
        s = Stream() 
        if n.isChord: n = n.chord21
        else: s.append(n.note21)
        sp = StreamPlayer(s)
        sp.play()
#        if n.isChord: 
#            s.append(n)
#        else: 
#            nn = Note(n.nameWithOctave)
#            s.append(nn)
#        sp = StreamPlayer(s)
#        sp.play()

      
    ##############################################
    def nameof(self, n):
        a = n.name+str(n.octave)
        if "-" in a:
            b = a.replace("C-","B")
            b = b.replace("D-","C#")
            b = b.replace("E-","D#")
            b = b.replace("F-","E")
            b = b.replace("G-","F#")
            b = b.replace("A-","G#")
            b = b.replace("B-","A#")
            return b 
        elif "E#" in a:
            b = a.replace("E#","F")
            return b         
        elif "B#" in a:
            b = a.replace("B#","C")
            return b         
        else:
            return a
    
    
    #####################################################################
    def playKeyboard(self):
        
        print "\nPlaying now at speed:", self.rate
        print "Right-mouse click to move 3D scene."
        print "Press <esc> at any time to quit."

        tR,tL = 9999,9999

        if self.rightHand:
            handR   = self.rightHand
            vphandR = self.vpRH
             
            engagedkeysR    = [False]*len(handR.noteseq)
            engagedfingersR = [False]*6 # element 0 is dummy
        
            tR    = handR.noteseq[0].time #start time of first note
            fineR = handR.noteseq[-1].time #end time

        if self.leftHand:
            handL   = self.leftHand
            vphandL = self.vpLH
             
            engagedkeysL    = [False]*len(handL.noteseq)
            engagedfingersL = [False]*6 # element 0 is dummy
        
            tL    = handL.noteseq[0].time #start time of first note
            fineL = handL.noteseq[-1].time #end time

        t = min(tR,tL)
        dt = 0.01
        txs=[]
        
        while True:
            
            if self.rightHand:
                for i, n in enumerate(handR.noteseq):##################### RIGHT
                    start, stop, f = n.time, n.time+n.duration, n.fingering
                    if stop <= t <= stop+dt and engagedkeysR[i]:  ###release key
                        engagedkeysR[i]    = False
                        engagedfingersR[f] = False
    
                        name = self.nameof(n)
                        ts   = self.KB[name]
                        #print t, '\t..release', name
                        ts.rotate(angle=-0.15)
                        vphandR[f].rotate(angle= +0.2, origin=(vphandR[f].x, 3.5, 7.5))
                        if "#" in name: ts.color = vp.color.black
                        else: ts.color = vp.color.white
                
                if t>fineR+dt: break
                
                for i, n in enumerate(handR.noteseq):##################### RIGHT
                    start, stop, f = n.time, n.time+n.duration, n.fingering
    
                    ### press key n with finger f
                    if start <= t < stop and not engagedkeysR[i] and not engagedfingersR[f]: 
                        engagedkeysR[i]    = True
                        engagedfingersR[f] = True
    
                        name = self.nameof(n)
                        ts   = self.KB[name]
                        #print t, '\tpressing', name, start,'->', stop, ' \tfinger=', f
                        
                        if i>=len(handR.fingerseq): return                    
                        for g in [1,2,3,4,5]: vphandR[g].x = handR.fingerseq[i][g]
                        vphandR[f].rotate(angle= -0.2, origin=(vphandR[f].x, 3.5, 7.5))
                        vphandR[0].x = vphandR[3].x+0.2 # index 0 is palm
                        ts.rotate(angle= +0.15)                   ### press key
                        ts.color = (0.604,0.808,0.875)#154 206 223
                        tx=vp.label(pos=(vphandR[f].x, 3, -9),text=name, opacity=0,
                                    height=10, xoffset=86-vphandR[f].x, yoffset=85, box=0, border=8)
                        txs.append(tx)
                        
                        if self.playsounds: self.playsound(n)
            
            ##########################################################            
            if self.leftHand:
                for i, n in enumerate(handL.noteseq):##################### LEFT
                    start, stop, f = n.time, n.time+n.duration, n.fingering
                    if stop <= t <= stop+dt and engagedkeysL[i]:  ###release key
                        engagedkeysL[i]    = False
                        engagedfingersL[f] = False
    
                        name = self.nameof(n)
                        ts   = self.KB[name]
                        #print t, '\t..release', name
                        ts.rotate(angle=-0.15)
                        vphandL[f].rotate(angle= +0.2, origin=(vphandL[f].x, 1.5, 7.5))
                        if "#" in name: ts.color = vp.color.black
                        else: ts.color = vp.color.white
                
                if t>fineL+dt: break
                
                for i, n in enumerate(handL.noteseq):##################### LEFT
                    start, stop, f = n.time, n.time+n.duration, n.fingering
    
                    ### press key n with finger f
                    if start <= t < stop and not engagedkeysL[i] and not engagedfingersL[f]: 
                        engagedkeysL[i]    = True
                        engagedfingersL[f] = True
    
                        name = self.nameof(n)
                        ts   = self.KB[name]
                        #print t, '\tpressing', name, start,'->', stop, ' \tfinger=', f
                        
                        if i>=len(handL.fingerseq): return
                        for g in [5,4,3,2,1]: vphandL[g].x = -handL.fingerseq[i][g]
                        vphandL[f].rotate(angle= -0.2, origin=(vphandL[f].x, 1.5, 7.5))
                        vphandL[0].x = vphandL[3].x+0.2 # index 0 is palm
                        ts.rotate(angle= +0.15)                   ### press key
                        ts.color = (0.604,0.808,0.875)#154 206 223
                        tx=vp.label(pos=(vphandL[f].x, 3, -9),text=name, opacity=0,
                                    height=10, xoffset=86-vphandL[f].x, yoffset=85, box=0, border=8)
                        txs.append(tx)
                        
                        if self.playsounds and not self.rightHand: self.playsound(n)
            
            for i,tt in enumerate(txs):
                tt.xoffset -= 1.5
                if tt.xoffset < -200: 
                    tt.visible = False
                    del tt
                    del txs[i]
                
            t += dt #time flows
            vp.rate(self.rate)


#################################################################### test
if __name__ == "__main__":

        vk=VirtualKeyboard()
        vk.build_RH(None)
        vk.build_LH(None)
#        from music21.note import Note
#        n = Note('A-', quarterLength=1.5)
#        vk.playsound(n)







#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:         VirtualKeyboard
# Purpose:      Find optimal fingering for piano scores
# URL:          https://github.com/marcomusy/pianoplayer
# Author:       Marco Musy
#-------------------------------------------------------------------------------
from __future__ import division, print_function
from music21.midi.realtime import StreamPlayer
from music21.stream import Stream
try:
    import vtk
    import vtkplotter
except:
    print("VirtualKeyboard: cannot find vtk or vtkplotter packages. Not installed?")
    print('Try: (sudo) pip install --upgrade vtkplotter')
    quit()


version = 'PianoPlayer 1.0.0'


###########################################################
class VirtualKeyboard:

    def __init__(self):
                  
        self.rate = 100
        
        self.KB = dict()
        self.vp = None
        
        self.rightHand = None
        self.leftHand  = None
                
        self.playsounds = False

        self.vpRH=None
        self.vpLH=None
        self.build_keyboard() 

    #######################################################
    def build_RH(self, hand):
    
        print('Building Right Hand..')
        self.rightHand = hand
        a1, a2, a3 = (0.8, 0, 0), (0, 1, 0), (0, 0, 5)
        f1 = self.vp.ellipsoid(pos=(-2.,1.5,6.2), axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
        f2 = self.vp.ellipsoid(pos=(-1.,1.5,4.2), axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
        f3 = self.vp.ellipsoid(pos=( 0.,1.5,4),   axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
        f4 = self.vp.ellipsoid(pos=( 1.,1.5,4.1), axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
        f5 = self.vp.ellipsoid(pos=( 2.,1.5,4.5), axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))

        a1, a2, a3 = (9.6, 0, 0), (0, 2, 0), (0, 0, 6)
        palm = self.vp.ellipsoid(pos=(0,1.6,10), axis1=a1, axis2=a2, axis3=a3,  alpha=0.3, c=(.5, 0, 0))

        self.vpRH = [palm, f1,f2,f3,f4,f5]
        for limb in self.vpRH: 
            limb.rotateX(90)
            limb.x( limb.x()* 2.5 )
            limb.pos(limb.pos() + [16.5*5.+1, 0, 0] )
    
    #######################################################
    def build_LH(self, hand):
    
        print('Building Left Hand..')
        self.leftHand = hand
        a1, a2, a3 = (0.8, 0, 0), (0, 1, 0), (0, 0, 5)
        f1 = self.vp.ellipsoid(pos=(-2.,1.5,6.2), axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
        f2 = self.vp.ellipsoid(pos=(-1.,1.5,4.2), axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
        f3 = self.vp.ellipsoid(pos=( 0.,1.5,4),   axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
        f4 = self.vp.ellipsoid(pos=( 1.,1.5,4.1), axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
        f5 = self.vp.ellipsoid(pos=( 2.,1.5,4.5), axis1=a1, axis2=a2, axis3=a3, alpha=0.9, c=(.7,0.3,0.3))
 
        a1, a2, a3 = (9.6, 0, 0), (0, 2, 0), (0, 0, 6)
        palm = self.vp.ellipsoid(pos=(0,1.6,10), axis1=a1, axis2=a2, axis3=a3,  alpha=0.3, c=(.5, 0, 0))

        self.vpLH = [palm, f1,f2,f3,f4,f5]
        for limb in self.vpLH: 
            limb.rotateX(90)
            limb.rotateY(180)
            limb.x( limb.x()* 2.5 )
            limb.pos(limb.pos() + [16.5*3.+1, 0, 3] )
                

    #######################################################
    def build_keyboard(self):
        
        print('Building Keyboard..')
        nts = ("C","D","E","F","G","A","B")
        tol = 0.12
        keybsize = 16.5 #cm, span of one octave
        wb = keybsize/7.
        nr_octaves = 7
        span = nr_octaves*wb*7.
    
        self.vp = vtkplotter.Plotter(title='Piano Keyboard', axes=0, size=(1200,600), bg='lb', verbose=0)

        #wooden top and base
        self.vp.box(pos=(span/2+keybsize, 6,  1), length=span+1, height=3, width= 5, texture='wood5') #top
        self.vp.box(pos=(span/2+keybsize, 0, -1), length=span+1, height=1, width=17, texture='wood5')
        self.vp.text(version, pos=(18,5.5,2.), depth=.7)
        #leggio
        leggio = self.vp.box(pos=(span/1.6,8,10), length=span/2, height=span/8, width=0.08, c=(1,1,0.9))
        leggio.rotateX(angle=-20)

        for ioct in range(nr_octaves):
            for ik in range(7): #white keys
                x  = ik * wb + (ioct+1.)*keybsize +wb/2.
                tb = self.vp.box(pos=(x,-2,0), length=wb-tol, height=1, width=12, c=(1,1,1))
                self.KB.update({nts[ik]+str(ioct+1) : tb})
                if not nts[ik] in ("E","B"): #black keys
                    tn=self.vp.box(pos=(x+wb/2,-0, 1), length=wb*.6, height=1, width=8, c=(0,0,0))
                    self.KB.update({nts[ik]+"#"+str(ioct+1) : tn})


    ####################################################### 
    def playsound(self, n):
        s = Stream() 
        if n.isChord: n = n.chord21
        else: s.append(n.note21)
        sp = StreamPlayer(s)
        sp.play()
#        if n.isChord: 
#            s.append(n)
#        else: 
#            nn = Note(n.nameWithOctave)
#            s.append(nn)
#        sp = StreamPlayer(s)
#        sp.play()
    
    
    #####################################################################
    def playKeyboard(self):
        
        print("\nPlaying now at speed:", self.rate)
        print("Right-mouse click to move 3D scene.")
        print("Press <esc> at any time to quit.")

        tR,tL = 9999,9999

        if self.rightHand:
            handR   = self.rightHand
            vphandR = self.vpRH
             
            engagedkeysR    = [False]*len(handR.noteseq)
            engagedfingersR = [False]*6 # element 0 is dummy
        
            tR    = handR.noteseq[0].time #start time of first note
            fineR = handR.noteseq[-1].time #end time

        if self.leftHand:
            handL   = self.leftHand
            vphandL = self.vpLH
             
            engagedkeysL    = [False]*len(handL.noteseq)
            engagedfingersL = [False]*6 # element 0 is dummy
        
            tL    = handL.noteseq[0].time #start time of first note
            fineL = handL.noteseq[-1].time #end time

        t = min(tR,tL)
        dt = 0.01
        
        while True:
            
            if self.rightHand:
                for i, n in enumerate(handR.noteseq):##################### RIGHT
                    start, stop, f = n.time, n.time+n.duration, n.fingering
                    if stop <= t <= stop+dt and engagedkeysR[i]:  ###release key
                        engagedkeysR[i]    = False
                        engagedfingersR[f] = False
    
                        name = nameof(n)
                        ts   = self.KB[name]
                        #print t, '\t..release', name
                        ts.rotateX(angle=-0.15)
                        vphandR[f].rotateX(angle= +0.2, origin=(vphandR[f].x, 3.5, 7.5))
                        #if "#" in name: ts.color = vp.color.black
                        #else: ts.color = vp.color.white
                
                if t>fineR+dt: break
                
                for i, n in enumerate(handR.noteseq):##################### RIGHT
                    start, stop, f = n.time, n.time+n.duration, n.fingering
    
                    ### press key n with finger f
                    if start <= t < stop and not engagedkeysR[i] and not engagedfingersR[f]: 
                        engagedkeysR[i]    = True
                        engagedfingersR[f] = True
    
                        name = nameof(n)
                        ts   = self.KB[name]
                        #print t, '\tpressing', name, start,'->', stop, ' \tfinger=', f
                        
                        if i>=len(handR.fingerseq): return                    
                        for g in [1,2,3,4,5]: vphandR[g].x = handR.fingerseq[i][g]
                        vphandR[f].rotateX(angle= -0.2, origin=(vphandR[f].x, 3.5, 7.5))
                        vphandR[0].x = vphandR[3].x+0.2 # index 0 is palm
                        ts.rotateX(angle= +0.15)                   ### press key
                        #ts.color = (0.604,0.808,0.875)#154 206 223
                        
                        if self.playsounds: self.playsound(n)
            
            ##########################################################            
            if self.leftHand:
                for i, n in enumerate(handL.noteseq):##################### LEFT
                    start, stop, f = n.time, n.time+n.duration, n.fingering
                    if stop <= t <= stop+dt and engagedkeysL[i]:  ###release key
                        engagedkeysL[i]    = False
                        engagedfingersL[f] = False
    
                        name = nameof(n)
                        ts   = self.KB[name]
                        #print t, '\t..release', name
                        ts.rotateX(angle=-0.15)
                        vphandL[f].rotateX(angle= +0.2, origin=(vphandL[f].x, 1.5, 7.5))
                        #if "#" in name: ts.color = vp.color.black
                        #else: ts.color = vp.color.white
                
                if t>fineL+dt: break
                
                for i, n in enumerate(handL.noteseq):##################### LEFT
                    start, stop, f = n.time, n.time+n.duration, n.fingering
    
                    ### press key n with finger f
                    if start <= t < stop and not engagedkeysL[i] and not engagedfingersL[f]: 
                        engagedkeysL[i]    = True
                        engagedfingersL[f] = True
    
                        name = nameof(n)
                        ts   = self.KB[name]
                        #print t, '\tpressing', name, start,'->', stop, ' \tfinger=', f
                        
                        if i>=len(handL.fingerseq): return
                        for g in [5,4,3,2,1]: vphandL[g].x = -handL.fingerseq[i][g]
                        vphandL[f].rotateX(angle= -0.2, origin=(vphandL[f].x, 1.5, 7.5))
                        vphandL[0].x = vphandL[3].x+0.2 # index 0 is palm
                        ts.rotateX(angle= +0.15)                   ### press key
                        #ts.color = (0.604,0.808,0.875)#154 206 223
                        
                        if self.playsounds and not self.rightHand: self.playsound(n)
                            
            t += dt #time flows

   
##############################################
def nameof(n):
    a = n.name+str(n.octave)
    if "-" in a:
        b = a.replace("C-","B")
        b = b.replace("D-","C#")
        b = b.replace("E-","D#")
        b = b.replace("F-","E")
        b = b.replace("G-","F#")
        b = b.replace("A-","G#")
        b = b.replace("B-","A#")
        return b 
    elif "E#" in a:
        b = a.replace("E#","F")
        return b         
    elif "B#" in a:
        b = a.replace("B#","C")
        return b         
    else:
        return a


#################################################################### test
if __name__ == "__main__":

    vk = VirtualKeyboard()
    vk.build_RH(None)
    vk.build_LH(None)

    vk.vp.show(zoom=2)

    # from music21.note import Note
    # n = Note('A-', quarterLength=1.5)
    # vk.playsound(n)







#
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 26 19:22:20 2015

@author: marco musy
"""
from __future__ import division, print_function

#####################################################
class INote:
    def __init__(self):
        self.note21   = None
        self.name     = None
        self.isChord  = False
        self.isBlack  = False
        self.octave   = 0
        self.x        = 0.0
        self.time     = 0.0
        self.duration = 0.0
        self.fingering= 0 
        self.measure  = 0
        self.chordnr  = 0


#####################################################
def keypos(n): #position of notes on keyboard
    step = 0.0
    keybsize = 16.5 #cm
    k = keybsize/7.0
    if   n.name == 'C'  : step = k *0.5
    elif n.name == 'D'  : step = k *1.5
    elif n.name == 'E'  : step = k *2.5
    elif n.name == 'F'  : step = k *3.5
    elif n.name == 'G'  : step = k *4.5
    elif n.name == 'A'  : step = k *5.5
    elif n.name == 'B'  : step = k *6.5
    elif n.name == 'B#' : step = k *0.5
    elif n.name == 'C#' : step = k
    elif n.name == 'D#' : step = k *2.
    elif n.name == 'E#' : step = k *3.5
    elif n.name == 'F#' : step = k *4.
    elif n.name == 'G#' : step = k *5.
    elif n.name == 'A#' : step = k *6.
    elif n.name == 'C-' : step = k *6.5
    elif n.name == 'D-' : step = k 
    elif n.name == 'E-' : step = k *2.
    elif n.name == 'F-' : step = k *2.5
    elif n.name == 'G-' : step = k *4.
    elif n.name == 'A-' : step = k *5.
    elif n.name == 'B-' : step = k *6.
    else: 
        print("ERROR note not found", n.name)
    return keybsize * n.octave + step
    

#####################################################
def reader(sf, beam=0):
    
    try:
        strm = sf.parts[beam].flat
    except AttributeError: 
        strm = sf.flat
    except IndexError:
        print('Beam nr.', beam, 'not found. Skip.')
        return []
    print('Reading beam', beam, 'made of', len(strm), 'objects...')

    noteseq = []

    for n in strm:

        if n.duration.quarterLength==0 : continue
        
        if n.tie and (n.tie.type=='continue' or n.tie.type=='stop'): continue

        if n.isNote:
            if len(noteseq) and n.offset == noteseq[-1].time:
                # print "doppia nota", n.name
                continue
            an        = INote()
            an.note21 = n
            an.isChord= False
            an.name   = n.name
            an.octave = n.octave
            an.measure= n.measureNumber
            an.x      = keypos(n)
            an.time   = n.offset
            an.duration = n.duration.quarterLength
            an.isBlack= False
            try:
                pc = n.pitch.pitchClass
            except AttributeError:
                pc = n.pitchClass
                an.isBlack = False
            if pc in [1, 3, 6, 8, 10]: an.isBlack = True
            if n.hasLyrics(): an.fingering = n.lyric
            noteseq.append(an)

        if n.isChord:
            if n.tie and (n.tie.type=='continue' or n.tie.type=='stop'): continue
            sfasam = 0.01 # sfasa leggermente le note dell'accordo
            for j, cn in enumerate(n.pitches):
                an = INote()
                an.isChord = True
                an.chord21 = n
                an.note21  = cn
                an.name    = cn.name
                an.chordnr = j
                an.octave  = cn.octave
                an.measure = n.measureNumber
                an.x       = keypos(cn)
                an.time    = n.offset                 +sfasam*j
                an.duration= n.duration.quarterLength -sfasam*(len(n.pitches)-1)
                try:
                    pc = cn.pitch.pitchClass
                except AttributeError:
                    pc = cn.pitchClass
                if pc in [1, 3, 6, 8, 10]: an.isBlack = True
                else: an.isBlack = False
                noteseq.append(an)

    print("Total nr of notes read:", len(noteseq))
    if len(noteseq)<2: 
        print("Beam is empty.")
        quit()    
    
    return noteseq




# -*- coding: utf-8 -*-
#-------------------------------------------------------------------------------
# Name:         PianoFing
# Purpose:      Find optimal fingering for piano scores
# Author:       Marco Musy
#-------------------------------------------------------------------------------
from __future__ import division, print_function

#####################################################
class Hand:
    def __init__(self, left_or_right="right", size='M'):
        
        self.LR        = left_or_right
        self.handspan  = None # thumb to pinkie distance in cm
        self.usable_fingers = (1, 2, 3, 4, 5)
        self.fpos      = [None,  -7.0,-2.8, 0.0, 2.8, 5.8] #first is dummy, (cm)
        self.weights   = [None,   1.1, 1.0, 1.1, 0.9, 0.8] #finger rel strenght
        self.bfactor   = [None,   0.5, 1.0, 1.1, 0.8, 0.6] #hit of black key bias
        self.noteseq   = []
        self.fingerseq = []
        self.depth     = 8
        self.autodepth = True  
        self.fstep     = 2 #recalculate fingering at each fstep
        
        self.setHandSize(size)
        if self.LR is not "right" and self.LR is not "left":
            print('Error: Hand must be left or right. Abort.', left_or_right)
            quit()

    #####################################################
    def setHandSize(self, s):
        self.size = s
        if    s=='M'  : f = 1.0
        elif  s=='XXS': f = 0.46
        elif  s=='XS' : f = 0.64
        elif  s=='S'  : f = 0.82
        elif  s=='L'  : f = 1.1
        elif  s=='XL' : f = 1.2
        elif  s=='XXL': f = 1.3
        else:
            print("Unknown hand size", s, "... Abort.")
            quit()
        
        for i in range(1,6): 
            if self.fpos[i]: self.fpos[i] *= f
        self.handspan = 21.0*f
        print('\nYour hand span set to size-'+s, 'which is',self.handspan,'cm')
        print('(max relaxed distance btw thumb and pinkie)\n')

    
    #####################################################
    def skiprules(self, fa,fb, na,nb): #two consecutive notes movement
        # fa is fingering for note na
        xba = nb.x - na.x #physical distance, cm

        #skip if repeat different note nb w/ same finger
        if fb==fa and xba!=0 and na.duration<4: return True 
    
        if fa>1 : ##if not thumb
            if fb>1  and (fb-fa)*xba<0:        
                return True #skip if non-thumb fingers are crossings
            if fb==1 and nb.isBlack and xba>0: 
                return True #crossing thumb goes to black key, skip.
        if na.isChord and nb.isChord and nb.time-na.time<4:
            if   fa >2 and abs(fb-fa)==1 and abs(xba)>4.5: 
                return True #dist max btw 2 consecutive fingers
            elif fa==1 and fb==2 and abs(xba)>12.: return True
            elif fa==2 and fb==1 and abs(xba)>12.: return True
            elif fa==2 and fb==3 and abs(xba)> 6.: return True
            elif fa==3 and fb==2 and abs(xba)> 6.: return True
#            if fa>1 and fb==1 and xba>0: return True #no thumb cross (not working?)
#            if fa==1 and fb>1 and xba<0: return True
        return False
    
    
    ##################################################### rigid hand
    def set_fingers_positions(self, fings, notes, i):
        fi = fings[i]
        ni = notes[i]
        newfpos = [0]*6
        ifx = self.fpos[fi]
        for j in self.usable_fingers:
            newfpos[j]= self.fpos[j] + ni.x - ifx
        return newfpos


    ##################################################### proportional stretching
    def set_fingers_positions_hide(self, fings, notes, i):
        fi = fings[i]
        nix = notes[i].x
        newfpos = [0]*6

        fac, minfac, maxfac = 1., 0.8, 1.4
        if i<len(notes)-1:
            fj = fings[i+1] #next note
            if fj :
                nj = notes[i+1]
                restd = self.fpos[fi]-self.fpos[fj] #rest distance btw 2 fings
                actud = nix - nj.x                  #actual distance btw 2 notes
                if restd: 
                    tfac  = abs(actud/restd)
                    if    tfac < minfac: fac = minfac   #allow whole hand stretching
                    elif  tfac > maxfac: fac = maxfac
                    else: fac = tfac 

        for j in self.usable_fingers:
            newfpos[j] = (self.fpos[j]-self.fpos[fi]) *fac + nix

        return newfpos


    #####################################################
    def ave_velocity(self, fingering, notes):
        ###calculate v for playing notes in fingering combination

        #initial position of first note played with finger fing[0]
        newfpos = self.set_fingers_positions(fingering, notes, 0)
        
        val = 0.
        for i in range(1, self.depth):
            na = notes[i-1]
            nb = notes[i]
            fb = fingering[i]
                        
            dx = nb.x - newfpos[fb]      #spazio percorso dal dito
            dt = nb.time - na.time +0.01 #tempo disponibile
            v  = dx/dt                   #velocita
            
            if nb.isBlack: bfac = self.bfactor[fb]
            else: bfac=1.
            v /= self.weights[fb] * bfac #penalty (increase speed)
            val += abs(v)
            
            #update all fingers positions
            newfpos = self.set_fingers_positions(fingering, notes, i)

        return val / (self.depth-1)*10.
        
        
    #####################################################
    #generate meaningful fingerings for note sequence 
    def optimize_seq(self, n, istart):
    
        if self.autodepth:
            for i in range(4,9):
                self.depth = i+1
                if n[i].time - n[0].time > 4: break #depth limit in secs
    
        n1, n2, n3, n4, n5 = n[0], n[1], n[2], n[3], n[4] 
        n6, n7, n8, n9 = [None]*4
        if self.depth>5: n6 = n[5]
        if self.depth>6: n7 = n[6]
        if self.depth>7: n8 = n[7]
        if self.depth>8: n9 = n[8]
        if istart == 0: u1 = self.usable_fingers
        else: u1 = [istart]
    
        bestval = 1.e+10
        bestcomb = None
        for f1 in u1:
            for f2 in self.usable_fingers:
                if self.skiprules(f1,f2,n1,n2): continue
                for f3 in self.usable_fingers:
                    if self.skiprules(f2,f3,n2,n3): continue
                    for f4 in self.usable_fingers:
                        if self.skiprules(f3,f4,n3,n4): continue
                        for f5 in self.usable_fingers:
                            if self.skiprules(f4,f5,n4,n5): continue
                            if self.depth<6: u=[False]
                            else: u=self.usable_fingers
                            for f6 in u:
                                if f6 and self.skiprules(f5,f6,n5,n6): continue
                                if self.depth<7: u=[False]
                                else: u=self.usable_fingers
                                for f7 in u:
                                    if f7 and self.skiprules(f6,f7,n6,n7): continue
                                    if self.depth<8: u=[False]
                                    else: u=self.usable_fingers
                                    for f8 in u:
                                        if f8 and self.skiprules(f7,f8,n7,n8): continue
                                        if self.depth<9: u=[False]
                                        else: u=self.usable_fingers
                                        for f9 in u:
                                            if f9 and self.skiprules(f8,f9,n8,n9): continue
                                            c = [f1,f2,f3,f4,f5,f6,f7,f8,f9]
                                            val = self.ave_velocity(c, n)
                                            if val < bestval: 
                                                bestcomb = c
                                                bestval  = val
        
        if bestcomb is None: return ([0]*self.depth, -1)       
        return (bestcomb, bestval)

    
    ###########################################################################################
    def generateFingering(self, nmeasures=1000):
    
        if self.LR == "left":
            #play left hand on a mirrored keyboard
            for anote in self.noteseq:
                anote.x = -anote.x 

        start_finger, out, costf = 0, [0]*9, 0

        for inote in range(len(self.noteseq)):
            an = self.noteseq[inote]
            i  = inote%self.fstep # position inside the group of step notes
            if an.measure > nmeasures : break
            
            if i==0:
                if inote <= len(self.noteseq)-9: 
                    ninenotes  = self.noteseq[inote : inote+9] 
                    out, costf = self.optimize_seq(ninenotes, start_finger)
                    best_finger= out[i]
                else: #close to the end of score
                    best_finger= out[9-len(self.noteseq) + inote]
            else:
                best_finger = out[i] 
            start_finger = out[i+1]
            
            bestfpos = self.set_fingers_positions(out, ninenotes, i)
            self.fingerseq.append(bestfpos)
            
            an.fingering = best_finger
            if best_finger>0:
                if an.isChord:
                    if not an.chord21.hasLyrics():
                        for li in range(len(an.chord21.pitches)): an.chord21.addLyric('0')
                    #an.chord21.addLyric(best_finger)
                    nl = len(an.chord21.pitches)-an.chordnr
                    an.chord21.addLyric(best_finger, nl)
                else:
                    an.note21.addLyric(best_finger)

            
            #-----------------------------

            print("meas."+str(an.measure), end='')
            print("  finger:"+str(best_finger) + " on " + an.name+str(an.octave), end='')
            if i==0:
                print("\tv=" + str(int(costf*10.)/10.), end='')
                if self.autodepth:
                    print("\t"+"   "+str(out[0:self.depth]) + " d =",self.depth)
                else:
                    print("\t"+("   "*(inote%self.depth))+str(out[0:self.depth]))
            else:
                print()
